@@ -9,40 +9,97 @@
     :license: BSD, see LICENSE for more details
 """
 
-__version__ = '0.9.1'
+__version__ = '0.13'
 __versionfull__ = __version__
 
-import uuid
+import base64
+import functools
 import hashlib
 import inspect
-import exceptions
-import functools
-
-from types import NoneType
+import logging
+import string
+import uuid
+import warnings
 
 from werkzeug import import_string
-from flask import request
+from flask import request, current_app
 
-JINJA_CACHE_ATTR_NAME = '_template_fragment_cache'
+from ._compat import PY2
+
+logger = logging.getLogger(__name__)
+
+TEMPLATE_FRAGMENT_KEY_TEMPLATE = '_template_fragment_cache_%s%s'
+
+# Used to remove control characters and whitespace from cache keys.
+valid_chars = set(string.ascii_letters + string.digits + '_.')
+delchars = ''.join(c for c in map(chr, range(256)) if c not in valid_chars)
+if PY2:
+    null_control = (None, delchars)
+else:
+    null_control = (dict((k,None) for k in delchars),)
 
 def function_namespace(f, args=None):
     """
     Attempts to returns unique namespace for function
     """
     m_args = inspect.getargspec(f)[0]
+    instance_token = None
 
-    if len(m_args) and args:
-        if m_args[0] == 'self':
-            return '%s.%s.%s' % (f.__module__, args[0].__class__.__name__, f.__name__)
-        elif m_args[0] == 'cls':
-            return '%s.%s.%s' % (f.__module__, args[0].__name__, f.__name__)
+    instance_self = getattr(f, '__self__', None)
 
-    if hasattr(f, 'im_func'):
-        return '%s.%s.%s' % (f.__module__, f.im_class.__name__, f.__name__)
-    elif hasattr(f, '__class__'):
-        return '%s.%s.%s' % (f.__module__, f.__class__.__name__, f.__name__)
+    if instance_self \
+    and not inspect.isclass(instance_self):
+        instance_token = repr(f.__self__)
+    elif m_args \
+    and m_args[0] == 'self' \
+    and args:
+        instance_token = repr(args[0])
+
+    module = f.__module__
+
+    if hasattr(f, '__qualname__'):
+        name = f.__qualname__
     else:
-        return '%s.%s' % (f.__module__, f.__name__)
+        klass = getattr(f, '__self__', None)
+
+        if klass \
+        and not inspect.isclass(klass):
+            klass = klass.__class__
+
+        if not klass:
+            klass = getattr(f, 'im_class', None)
+
+        if not klass:
+            if m_args and args:
+                if m_args[0] == 'self':
+                    klass = args[0].__class__
+                elif m_args[0] == 'cls':
+                    klass = args[0]
+
+        if klass:
+            name = klass.__name__ + '.' + f.__name__
+        else:
+            name = f.__name__
+
+    ns = '.'.join((module, name))
+    ns = ns.translate(*null_control)
+
+    if instance_token:
+        ins = '.'.join((module, name, instance_token))
+        ins = ins.translate(*null_control)
+    else:
+        ins = None
+
+    return ns, ins
+
+def make_template_fragment_key(fragment_name, vary_on=[]):
+    """
+    Make a cache key for a specific fragment name
+    """
+    if vary_on:
+        fragment_name = "%s_" % fragment_name
+    return TEMPLATE_FRAGMENT_KEY_TEMPLATE % (fragment_name, "_".join(vary_on))
+
 
 #: Cache Object
 ################
@@ -53,63 +110,86 @@ class Cache(object):
     """
 
     def __init__(self, app=None, with_jinja2_ext=True, config=None):
+        if not (config is None or isinstance(config, dict)):
+            raise ValueError("`config` must be an instance of dict or None")
+
         self.with_jinja2_ext = with_jinja2_ext
         self.config = config
 
-        self.cache = None
-
+        self.app = app
         if app is not None:
-            self.init_app(app)
-        else:
-            self.app = None
-
-        self._memoized = []
+            self.init_app(app, config)
 
     def init_app(self, app, config=None):
         "This is used to initialize cache with your app object"
+        if not (config is None or isinstance(config, dict)):
+            raise ValueError("`config` must be an instance of dict or None")
 
-        if config is not None:
-            self.config = config
-        elif self.config is None:
-            self.config = app.config
+        #: Ref PR #44.
+        #: Do not set self.app in the case a single instance of the Cache
+        #: object is being used for multiple app instances.
+        #: Example use case would be Cache shipped as part of a blueprint
+        #: or utility library.
 
-        if not isinstance(self.config, (NoneType, dict)):
-            raise ValueError("`config` must be an instance of dict or NoneType")
+        base_config = app.config.copy()
+        if self.config:
+            base_config.update(self.config)
+        if config:
+            base_config.update(config)
+        config = base_config
 
-        self.config.setdefault('CACHE_DEFAULT_TIMEOUT', 300)
-        self.config.setdefault('CACHE_THRESHOLD', 500)
-        self.config.setdefault('CACHE_KEY_PREFIX', None)
-        self.config.setdefault('CACHE_MEMCACHED_SERVERS', None)
-        self.config.setdefault('CACHE_DIR', None)
-        self.config.setdefault('CACHE_OPTIONS', None)
-        self.config.setdefault('CACHE_ARGS', [])
-        self.config.setdefault('CACHE_TYPE', 'null')
+        config.setdefault('CACHE_DEFAULT_TIMEOUT', 300)
+        config.setdefault('CACHE_THRESHOLD', 500)
+        config.setdefault('CACHE_KEY_PREFIX', 'flask_cache_')
+        config.setdefault('CACHE_MEMCACHED_SERVERS', None)
+        config.setdefault('CACHE_DIR', None)
+        config.setdefault('CACHE_OPTIONS', None)
+        config.setdefault('CACHE_ARGS', [])
+        config.setdefault('CACHE_TYPE', 'null')
+        config.setdefault('CACHE_NO_NULL_WARNING', False)
+
+        if config['CACHE_TYPE'] == 'null' and not config['CACHE_NO_NULL_WARNING']:
+            warnings.warn("Flask-Cache: CACHE_TYPE is set to null, "
+                          "caching is effectively disabled.")
 
         if self.with_jinja2_ext:
-            setattr(app.jinja_env, JINJA_CACHE_ATTR_NAME, self)
+            from .jinja2ext import CacheExtension, JINJA_CACHE_ATTR_NAME
 
-            from flask.ext.cache.jinja2ext import CacheExtension
+            setattr(app.jinja_env, JINJA_CACHE_ATTR_NAME, self)
             app.jinja_env.add_extension(CacheExtension)
 
-        self.app = app
+        self._set_cache(app, config)
 
-        self._set_cache()
-
-    def _set_cache(self):
-        import_me = self.config['CACHE_TYPE']
+    def _set_cache(self, app, config):
+        import_me = config['CACHE_TYPE']
         if '.' not in import_me:
-            import_me = 'flask.ext.cache.backends.' + \
-                        import_me
+            from . import backends
 
-        cache_obj = import_string(import_me)
-        cache_args = self.config['CACHE_ARGS'][:]
-        cache_options = dict(default_timeout= \
-                             self.config['CACHE_DEFAULT_TIMEOUT'])
+            try:
+                cache_obj = getattr(backends, import_me)
+            except AttributeError:
+                raise ImportError("%s is not a valid FlaskCache backend" % (
+                                  import_me))
+        else:
+            cache_obj = import_string(import_me)
 
-        if self.config['CACHE_OPTIONS']:
-            cache_options.update(self.config['CACHE_OPTIONS'])
+        cache_args = config['CACHE_ARGS'][:]
+        cache_options = {'default_timeout': config['CACHE_DEFAULT_TIMEOUT']}
 
-        self.cache = cache_obj(self.app, self.config, cache_args, cache_options)
+        if config['CACHE_OPTIONS']:
+            cache_options.update(config['CACHE_OPTIONS'])
+
+        if not hasattr(app, 'extensions'):
+            app.extensions = {}
+
+        app.extensions.setdefault('cache', {})
+        app.extensions['cache'][self] = cache_obj(
+                app, config, cache_args, cache_options)
+
+    @property
+    def cache(self):
+        app = self.app or current_app
+        return app.extensions['cache'][self]
 
     def get(self, *args, **kwargs):
         "Proxy function for internal cache object."
@@ -131,6 +211,10 @@ class Cache(object):
         "Proxy function for internal cache object."
         self.cache.delete_many(*args, **kwargs)
 
+    def clear(self):
+        "Proxy function for internal cache object."
+        self.cache.clear()
+
     def get_many(self, *args, **kwargs):
         "Proxy function for internal cache object."
         return self.cache.get_many(*args, **kwargs)
@@ -200,13 +284,25 @@ class Cache(object):
                 if callable(unless) and unless() is True:
                     return f(*args, **kwargs)
 
-                cache_key = decorated_function.make_cache_key(*args, **kwargs)
+                try:
+                    cache_key = decorated_function.make_cache_key(*args, **kwargs)
+                    rv = self.cache.get(cache_key)
+                except Exception:
+                    if current_app.debug:
+                        raise
+                    logger.exception("Exception possibly due to cache backend.")
+                    return f(*args, **kwargs)
 
-                rv = self.get(cache_key)
                 if rv is None:
                     rv = f(*args, **kwargs)
-                    self.cache.set(cache_key, rv,
+                    try:
+                        self.cache.set(cache_key, rv,
                                    timeout=decorated_function.cache_timeout)
+                    except Exception:
+                        if current_app.debug:
+                            raise
+                        logger.exception("Exception possibly due to cache backend.")
+                        return f(*args, **kwargs)
                 return rv
 
             def make_cache_key(*args, **kwargs):
@@ -217,8 +313,6 @@ class Cache(object):
                 else:
                     cache_key = key_prefix
 
-                cache_key = cache_key.encode('utf-8')
-
                 return cache_key
 
             decorated_function.uncached = f
@@ -231,24 +325,60 @@ class Cache(object):
     def _memvname(self, funcname):
         return funcname + '_memver'
 
-    def memoize_make_version_hash(self):
-        return uuid.uuid4().bytes.encode('base64')[:6]
+    def _memoize_make_version_hash(self):
+        return base64.b64encode(uuid.uuid4().bytes)[:6].decode('utf-8')
 
-    def memoize_make_cache_key(self, make_name=None):
+    def _memoize_version(self, f, args=None,
+                         reset=False, delete=False, timeout=None):
+        """
+        Updates the hash version associated with a memoized function or method.
+        """
+        fname, instance_fname = function_namespace(f, args=args)
+        version_key = self._memvname(fname)
+        fetch_keys = [version_key]
+
+        if instance_fname:
+            instance_version_key = self._memvname(instance_fname)
+            fetch_keys.append(instance_version_key)
+
+        # Only delete the per-instance version key or per-function version
+        # key but not both.
+        if delete:
+            self.cache.delete_many(fetch_keys[-1])
+            return fname, None
+
+        version_data_list = list(self.cache.get_many(*fetch_keys))
+        dirty = False
+
+        if version_data_list[0] is None:
+            version_data_list[0] = self._memoize_make_version_hash()
+            dirty = True
+
+        if instance_fname and version_data_list[1] is None:
+            version_data_list[1] = self._memoize_make_version_hash()
+            dirty = True
+
+        # Only reset the per-instance version or the per-function version
+        # but not both.
+        if reset:
+            fetch_keys = fetch_keys[-1:]
+            version_data_list = [self._memoize_make_version_hash()]
+            dirty = True
+
+        if dirty:
+            self.cache.set_many(dict(zip(fetch_keys, version_data_list)),
+                                timeout=timeout)
+
+        return fname, ''.join(version_data_list)
+
+    def _memoize_make_cache_key(self, make_name=None, timeout=None):
         """
         Function used to create the cache_key for memoized functions.
         """
         def make_cache_key(f, *args, **kwargs):
-            fname = function_namespace(f, args)
-
-            version_key = self._memvname(fname)
-            version_data = self.get(version_key)
-
-            if version_data is None:
-                version_data = self.memoize_make_version_hash()
-                self.cache.set(version_key, version_data)
-
-            cache_key = hashlib.md5()
+            _timeout = getattr(timeout, 'cache_timeout', timeout)
+            fname, version_data = self._memoize_version(f, args=args,
+                                                        timeout=_timeout)
 
             #: this should have to be after version_data, so that it
             #: does not break the delete_memoized functionality.
@@ -258,21 +388,27 @@ class Cache(object):
                 altfname = fname
 
             if callable(f):
-                args, kwargs = self.memoize_kwargs_to_args(f, *args, **kwargs)
+                keyargs, keykwargs = self._memoize_kwargs_to_args(f,
+                                                                 *args,
+                                                                 **kwargs)
+            else:
+                keyargs, keykwargs = args, kwargs
 
             try:
-                updated = "{0}{1}{2}".format(altfname, args, kwargs)
+                updated = "{0}{1}{2}".format(altfname, keyargs, keykwargs)
             except AttributeError:
-                updated = "%s%s%s" % (altfname, args, kwargs)
+                updated = "%s%s%s" % (altfname, keyargs, keykwargs)
 
-            cache_key.update(updated)
-            cache_key = cache_key.digest().encode('base64')[:16]
+            cache_key = hashlib.md5()
+            cache_key.update(updated.encode('utf-8'))
+            cache_key = base64.b64encode(cache_key.digest())[:16]
+            cache_key = cache_key.decode('utf-8')
             cache_key += version_data
 
             return cache_key
         return make_cache_key
 
-    def memoize_kwargs_to_args(self, f, *args, **kwargs):
+    def _memoize_kwargs_to_args(self, f, *args, **kwargs):
         #: Inspect the arguments to the function
         #: This allows the memoization to be the same
         #: whether the function was called with
@@ -281,20 +417,25 @@ class Cache(object):
         arg_num = 0
         argspec = inspect.getargspec(f)
 
-        for i in range(len(argspec.args)):
+        args_len = len(argspec.args)
+        for i in range(args_len):
             if i == 0 and argspec.args[i] in ('self', 'cls'):
-                #: use the id of the class instance
+                #: use the repr of the class instance
                 #: this supports instance methods for
-                #: the memoized functions.
-                arg = id(args[0])
+                #: the memoized functions, giving more
+                #: flexibility to developers
+                arg = repr(args[0])
                 arg_num += 1
             elif argspec.args[i] in kwargs:
                 arg = kwargs[argspec.args[i]]
             elif arg_num < len(args):
                 arg = args[arg_num]
                 arg_num += 1
+            elif abs(i-args_len) <= len(argspec.defaults):
+                arg = argspec.defaults[i-args_len]
+                arg_num += 1
             else:
-                arg = argspec.defaults[i]
+                arg = None
                 arg_num += 1
 
             #: Attempt to convert all arguments to a
@@ -383,18 +524,30 @@ class Cache(object):
                 if callable(unless) and unless() is True:
                     return f(*args, **kwargs)
 
-                cache_key = decorated_function.make_cache_key(f, *args, **kwargs)
+                try:
+                    cache_key = decorated_function.make_cache_key(f, *args, **kwargs)
+                    rv = self.cache.get(cache_key)
+                except Exception:
+                    if current_app.debug:
+                        raise
+                    logger.exception("Exception possibly due to cache backend.")
+                    return f(*args, **kwargs)
 
-                rv = self.get(cache_key)
                 if rv is None:
                     rv = f(*args, **kwargs)
-                    self.cache.set(cache_key, rv,
+                    try:
+                        self.cache.set(cache_key, rv,
                                    timeout=decorated_function.cache_timeout)
+                    except Exception:
+                        if current_app.debug:
+                            raise
+                        logger.exception("Exception possibly due to cache backend.")
                 return rv
 
             decorated_function.uncached = f
             decorated_function.cache_timeout = timeout
-            decorated_function.make_cache_key = self.memoize_make_cache_key(make_name)
+            decorated_function.make_cache_key = self._memoize_make_cache_key(
+                                                make_name, decorated_function)
             decorated_function.delete_memoized = lambda: self.delete_memoized(f)
 
             return decorated_function
@@ -404,7 +557,7 @@ class Cache(object):
         """
         Deletes the specified functions caches, based by given parameters.
         If parameters are given, only the functions that were memoized with them
-        will be erased. Otherwise all the versions of the caches will be deleted.
+        will be erased. Otherwise all versions of the caches will be forgotten.
 
         Example::
 
@@ -437,6 +590,40 @@ class Cache(object):
             >>> param_func(2, 2)
             47
 
+        Delete memoized is also smart about instance methods vs class methods.
+
+        When passing a instancemethod, it will only clear the cache related
+        to that instance of that object. (object uniqueness can be overridden
+            by defining the __repr__ method, such as user id).
+
+        When passing a classmethod, it will clear all caches related across
+        all instances of that class.
+
+        Example::
+
+            class Adder(object):
+                @cache.memoize()
+                def add(self, b):
+                    return b + random.random()
+
+        .. code-block:: pycon
+
+            >>> adder1 = Adder()
+            >>> adder2 = Adder()
+            >>> adder1.add(3)
+            3.23214234
+            >>> adder2.add(3)
+            3.60898509
+            >>> cache.delete_memoized(adder.add)
+            >>> adder1.add(3)
+            3.01348673
+            >>> adder2.add(3)
+            3.60898509
+            >>> cache.delete_memoized(Adder.add)
+            >>> adder1.add(3)
+            3.53235667
+            >>> adder2.add(3)
+            3.72341788
 
         :param fname: Name of the memoized function, or a reference to the function.
         :param \*args: A list of positional parameters used with memoized function.
@@ -449,7 +636,7 @@ class Cache(object):
             instead of the function name, Flask-Cache will be able to place
             the args/kwargs in the proper order, and delete the positional cache.
 
-            However, if ``delete_memozied`` is just called with the name of the
+            However, if ``delete_memoized`` is just called with the name of the
             function, be sure to pass in potential arguments in the same order
             as defined in your function as args only, otherwise Flask-Cache
             will not be able to compute the same cache key.
@@ -468,16 +655,40 @@ class Cache(object):
             results would eventually be reclaimed by the caching backend.
         """
         if not callable(f):
-            raise exceptions.DeprecationWarning("Deleting messages by relative name is no longer"
-                          " reliable, please switch to a function reference"
-                          " or use the full function import name")
+            raise DeprecationWarning("Deleting messages by relative name is no longer"
+                          " reliable, please switch to a function reference")
 
-        _fname = function_namespace(f, args)
 
-        if not args and not kwargs:
-            version_key = self._memvname(_fname)
-            version_data = self.memoize_make_version_hash()
-            self.cache.set(version_key, version_data)
-        else:
-            cache_key = f.make_cache_key(f.uncached, *args, **kwargs)
-            self.cache.delete(cache_key)
+        try:
+            if not args and not kwargs:
+                self._memoize_version(f, reset=True)
+            else:
+                cache_key = f.make_cache_key(f.uncached, *args, **kwargs)
+                self.cache.delete(cache_key)
+        except Exception:
+            if current_app.debug:
+                raise
+            logger.exception("Exception possibly due to cache backend.")
+
+    def delete_memoized_verhash(self, f, *args):
+        """
+        Delete the version hash associated with the function.
+
+        ..warning::
+
+            Performing this operation could leave keys behind that have
+            been created with this version hash. It is up to the application
+            to make sure that all keys that may have been created with this
+            version hash at least have timeouts so they will not sit orphaned
+            in the cache backend.
+        """
+        if not callable(f):
+            raise DeprecationWarning("Deleting messages by relative name is no longer"
+                          " reliable, please use a function reference")
+
+        try:
+            self._memoize_version(f, delete=True)
+        except Exception:
+            if current_app.debug:
+                raise
+            logger.exception("Exception possibly due to cache backend.")
